# Generated by Django 3.1.8 on 2021-04-07 06:27

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('paper_uploads', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='collection',
            options={'default_manager_name': 'default_mgr', 'default_permissions': ()},
        ),
        migrations.AlterModelOptions(
            name='fileitem',
            options={'base_manager_name': 'objects', 'default_permissions': (), 'verbose_name': 'File item', 'verbose_name_plural': 'File items'},
        ),
        migrations.AlterModelOptions(
            name='imagecollection',
            options={'default_permissions': ()},
        ),
        migrations.AlterModelOptions(
            name='imageitem',
            options={'base_manager_name': 'objects', 'default_permissions': (), 'verbose_name': 'Image item', 'verbose_name_plural': 'Image items'},
        ),
        migrations.AlterModelOptions(
            name='svgitem',
            options={'base_manager_name': 'objects', 'default_permissions': (), 'verbose_name': 'SVG item', 'verbose_name_plural': 'SVG items'},
        ),
    ]



from typing import Any, Type, cast

from django.core.files.uploadedfile import UploadedFile
from django.http import HttpResponse
from django.utils.module_loading import import_string

from ..helpers import run_validators
from ..models.base import FileResource
from ..models.mixins import BacklinkModelMixin, EditableResourceMixin
from . import helpers
from .base import ChangeFileViewBase, DeleteFileViewBase, UploadFileViewBase


class UploadFileView(UploadFileViewBase):
    def get_file_model(self) -> Type[FileResource]:
        content_type_id = self.request.POST.get("paperContentType")
        return helpers.get_model_class(content_type_id, FileResource)

    def get_instance(self) -> FileResource:
        file_model = self.get_file_model()
        return file_model(
            owner_app_label=self.request.POST.get("paperOwnerAppLabel"),
            owner_model_name=self.request.POST.get("paperOwnerModelName"),
            owner_fieldname=self.request.POST.get("paperOwnerFieldName"),
        )

    def handle(self, file: UploadedFile) -> HttpResponse:
        instance = self.get_instance()

        owner_field = None
        if isinstance(instance, BacklinkModelMixin):
            owner_field = instance.get_owner_field()

        instance.attach(file)

        try:
            instance.full_clean()
            if owner_field is not None:
                run_validators(file, owner_field.validators)
        except Exception:
            instance.delete_file()
            raise

        instance.save()

        if not file.closed:
            file.close()

        return self.success(instance)

    def success(self, instance: FileResource) -> HttpResponse:
        return self.success_response(instance.as_dict())


class DeleteFileView(DeleteFileViewBase):
    def get_file_model(self) -> Type[FileResource]:
        content_type_id = self.request.POST.get("paperContentType")
        return helpers.get_model_class(content_type_id, FileResource)

    def get_file_id(self) -> Any:
        return self.request.POST.get("pk")

    def get_instance(self) -> FileResource:
        file_model = self.get_file_model()
        file_id = self.get_file_id()
        return cast(FileResource, helpers.get_instance(file_model, file_id))

    def handle(self) -> HttpResponse:
        instance = self.get_instance()
        instance.delete()
        return self.success()

    def success(self) -> HttpResponse:
        return self.success_response()


class ChangeFileView(ChangeFileViewBase):
    template_name = "paper_uploads/dialogs/file.html"

    def get_file_model(self) -> Type[FileResource]:
        content_type_id = self.request.GET.get("paperContentType")
        return helpers.get_model_class(content_type_id, FileResource)

    def get_file_id(self) -> Any:
        return self.request.GET.get("pk")

    def get_instance(self) -> FileResource:
        file_model = self.get_file_model()
        file_id = self.get_file_id()
        return cast(FileResource, helpers.get_instance(file_model, file_id))

    def get_form_class(self):
        if self.form_class is not None:
            return self.form_class

        file_model = self.get_file_model()
        if issubclass(file_model, EditableResourceMixin):
            return import_string(file_model.change_form_class)



from django.template import library
from ..utils import filesizeformat

register = library.Library()


@register.filter
@register.filter(name="filesizeformat", is_safe=True)
def do_filesizeformat(bytes_):
    return filesizeformat(bytes_)



import django
from django.db.models.signals import Signal


if django.VERSION >= (4, 0):
    # изменение контрольной суммы файла
    checksum_update = Signal()

    # создание / перезапись (recut) файла вариации изображения
    variation_created = Signal()

    # присоединение файла к ресурсу
    pre_attach_file = Signal()
    post_attach_file = Signal()

    # переименование файла
    pre_rename_file = Signal()
    post_rename_file = Signal()

    # удаление файла
    pre_delete_file = Signal()
    post_delete_file = Signal()

    # изменен порядок элементов коллекции
    collection_reordered = Signal()
else:
    # изменение контрольной суммы файла
    checksum_update = Signal(providing_args=["instance", "checksum"])

    # создание / перезапись (recut) файла вариации изображения
    variation_created = Signal(providing_args=["instance", "name"])

    # присоединение файла к ресурсу
    pre_attach_file = Signal(providing_args=["instance", "file", "options"])
    post_attach_file = Signal(providing_args=["instance", "file", "options", "response"])

    # переименование файла
    pre_rename_file = Signal(providing_args=["instance", "old_name", "new_name", "options"])
    post_rename_file = Signal(providing_args=["instance", "old_name", "new_name", "options", "response"])

    # удаление файла
    pre_delete_file = Signal(providing_args=["instance", "options"])
    post_delete_file = Signal(providing_args=["instance", "options", "response"])

    # изменен порядок элементов коллекции
    collection_reordered = Signal(providing_args=["instance"])



from django.db.migrations import Migration


class ExtendableMigration:
    """
    Класс, который обходит все элементы списка операций миграции
    и на ходу добавляет новые операции.
    """

    def __init__(self, migration: Migration, backward: bool = False, **kwargs):
        self.migration = migration
        self.backward = backward
        self.iter_index = 0
        self.insert_index = 0
        self.kwargs = kwargs

    def iterate(self):
        while True:
            try:
                op = self.migration.operations[self.iter_index]
            except LookupError:
                break

            self.insert_index = self.iter_index
            self.process(op, **self.kwargs)
            self.iter_index += 1

    def process(self, operation, **kwargs):
        raise NotImplementedError

    def insert_before(self, operation):
        # Добавление операции перед текущей.
        if self.backward:
            self.insert_index += 1
            self.migration.operations.insert(
                self.insert_index,
                operation
            )
            self.iter_index += 1
        else:
            self.migration.operations.insert(
                self.insert_index,
                operation
            )
            self.iter_index += 1

    def insert_after(self, operation):
        # Добавление операции после текущей.
        if self.backward:
            self.migration.operations.insert(
                self.insert_index,
                operation
            )
            self.iter_index += 1
        else:
            self.insert_index += 1
            self.migration.operations.insert(
                self.insert_index,
                operation
            )
            self.iter_index += 1



# Generated by Django 3.2.9 on 2021-12-17 10:15

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('paper_uploads', '0007_alter_collection_options'),
    ]

    operations = [
        migrations.RenameField(
            model_name='collectionitembase',
            old_name='item_type',
            new_name='type',
        ),
    ]



import os
import warnings
from typing import Sequence, Union

import magic
from django.core.exceptions import ValidationError
from django.utils.deconstruct import deconstructible
from django.utils.translation import gettext_lazy as _
from PIL import Image

from .typing import FileLike
from .utils import filesizeformat, parse_filesize, remove_dulpicates

__all__ = [
    "ExtensionValidator",
    "MimeTypeValidator",
    "MaxSizeValidator",
    "SizeValidator",
    "ImageMinSizeValidator",
    "ImageMaxSizeValidator",
]


@deconstructible
class ExtensionValidator:
    message = _("File `%(name)s` has an invalid extension. Valid extension(s): %(allowed)s")
    code = "invalid_extension"

    def __init__(self, allowed: Sequence[str], message=None):
        self.allowed = remove_dulpicates(ext.lstrip(".").lower() for ext in allowed)
        if message:
            self.message = message

    def __call__(self, file: FileLike):
        _, ext = os.path.splitext(file.name)
        ext = ext.lstrip(".").lower()
        params = {
            "name": file.name,
            "ext": ext,
            "allowed": ", ".join(self.allowed),
        }
        if ext not in self.allowed:
            raise ValidationError(self.message, code=self.code, params=params)

    def get_help_text(self):
        return "{}: {}".format(_("Allowed extensions"), ", ".join(self.allowed))


@deconstructible
class MimeTypeValidator:
    message = _("File `%(name)s` has an invalid mimetype '%(mimetype)s'")
    code = "invalid_mimetype"

    def __init__(self, allowed: Sequence[str], message=None):
        self.allowed = remove_dulpicates(mime.lower() for mime in allowed)
        if message:
            self.message = message

    def __call__(self, file: FileLike):
        file.seek(0)  # ensure read from the start
        mimetype = magic.from_buffer(file.read(1024), mime=True)
        file.seek(0)  # correct file position after mimetype detection
        basetype, subtype = mimetype.split("/", 1)
        params = {
            "name": file.name,
            "mimetype": mimetype,
            "allowed": ", ".join(self.allowed),
        }
        if not (mimetype in self.allowed or "{}/*".format(basetype) in self.allowed):
            raise ValidationError(self.message, code=self.code, params=params)

    def get_help_text(self):
        return "{}: {}".format(_("Allowed types"), ", ".join(self.allowed))


@deconstructible
class MaxSizeValidator:
    message = _("File `%(name)s` is too large. Maximum file size is %(limit_value)s.")
    code = "size_limit"

    def __init__(self, limit_value: Union[int, str], message=None):
        if isinstance(limit_value, str):
            self.limit_value = parse_filesize(limit_value)
        else:
            self.limit_value = limit_value

        if message:
            self.message = message

    def __call__(self, file: FileLike):
        params = {
            "name": file.name,
            "size": file.size,
            "limit_value": filesizeformat(self.limit_value),
        }
        if file.size > self.limit_value:
            raise ValidationError(self.message, code=self.code, params=params)

    def get_help_text(self):
        return "{}: {}".format(_("Maximum file size"), filesizeformat(self.limit_value))


class SizeValidator(MaxSizeValidator):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            "SizeValidator is deprecated in favor of MaxSizeValidator",
            DeprecationWarning,
            stacklevel=2
        )
        super().__init__(*args, **kwargs)


@deconstructible
class ImageMinSizeValidator:
    error_messages = {
        "min_width": _(
            "Image `%(name)s` is not wide enough. The minimum width is %(width_limit)s pixels."
        ),
        "min_height": _(
            "Image `%(name)s` is not tall enough. The minimum height is %(height_limit)s pixels."
        ),
        "min_size": _(
            "Image `%(name)s` is too small. Image should be at least %(width_limit)sx%(height_limit)s pixels."
        ),
    }

    def __init__(self, width: int = 0, height: int = 0):
        self.width_limit = width
        self.height_limit = height

    def __call__(self, file: FileLike):
        if file.closed:
            raise ValidationError("File `%(name)s` is closed" % {"name": os.path.basename(file.name)})

        try:
            img = Image.open(file)
            image_size = img.size
        except OSError:
            raise ValidationError("File `%(name)s` is not an image" % {"name": os.path.basename(file.name)})

        params = {
            "width_limit": self.width_limit,
            "height_limit": self.height_limit,
            "size": image_size,
            "name": file.name,
        }

        invalid_width = self.width_limit and image_size[0] < self.width_limit
        invalid_height = self.height_limit and image_size[1] < self.height_limit
        if invalid_width:
            if invalid_height:
                code = "min_size"
            else:
                code = "min_width"
        elif invalid_height:
            code = "min_height"
        else:
            return

        raise ValidationError(self.error_messages[code], code=code, params=params)

    def get_help_text(self):
        if self.width_limit:
            if self.height_limit:
                return "{}: {}".format(
                    _("Minimum dimensions"),
                    _("%(min_width)sx%(min_height)s pixels") % {
                        "min_width": self.width_limit,
                        "min_height": self.height_limit
                    }
                )
            else:
                return "{}: {}".format(_("Minimum image width"), _("%s pixels") % self.width_limit)
        elif self.height_limit:
            return "{}: {}".format(_("Minimum image height"), _("%s pixels") % self.height_limit)


@deconstructible
class ImageMaxSizeValidator:
    error_messages = {
        "max_width": _(
            "Image `%(name)s` is too wide. The maximum width is %(width_limit)s pixels."
        ),
        "max_height": _(
            "Image `%(name)s` is too tall. The maximum height is %(height_limit)s pixels."
        ),
        "max_size": _(
            "Image `%(name)s` is too big. Image should be at most %(width_limit)sx%(height_limit)s pixels."
        ),
    }

    def __init__(self, width: int = 0, height: int = 0):
        self.width_limit = width
        self.height_limit = height

    def __call__(self, file: FileLike):
        if file.closed:
            raise ValidationError("File `%(name)s` is closed" % {"name": os.path.basename(file.name)})

        try:
            img = Image.open(file)
            image_size = img.size
        except OSError:
            raise ValidationError("File `%(name)s` is not an image" % {"name": os.path.basename(file.name)})

        params = {
            "width_limit": self.width_limit,
            "height_limit": self.height_limit,
            "size": image_size,
            "name": file.name,
        }

        invalid_width = self.width_limit and image_size[0] > self.width_limit
        invalid_height = self.height_limit and image_size[1] > self.height_limit
        if invalid_width:
            if invalid_height:
                code = "max_size"
            else:
                code = "max_width"
        elif invalid_height:
            code = "max_height"
        else:
            return

        raise ValidationError(self.error_messages[code], code=code, params=params)

    def get_help_text(self):
        if self.width_limit:
            if self.height_limit:
                return "{}: {}".format(
                    _("Maximum dimensions"),
                    _("%(max_width)sx%(max_height)s pixels") % {
                        "max_width": self.width_limit,
                        "max_height": self.height_limit
                    },
                )
            else:
                return "{}: {}".format(_("Maximum image width"), _("%s pixels") % self.width_limit)
        elif self.height_limit:
            return "{}: {}".format(_("Maximum image height"), _("%s pixels") % self.height_limit)



# Generated by Django 4.0.2 on 2022-03-25 10:15

from django.db import migrations, models
import django.utils.timezone
import paper_uploads.models.fields.base
import paper_uploads.models.mixins


class Migration(migrations.Migration):

    dependencies = [
        ('paper_uploads', '0006_svgitem_description_svgitem_title'),
    ]

    operations = [
        migrations.CreateModel(
            name='UploadedSVGFile',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('owner_app_label', models.CharField(editable=False, max_length=100)),
                ('owner_model_name', models.CharField(editable=False, max_length=100)),
                ('owner_fieldname', models.CharField(editable=False, max_length=255)),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='created at')),
                ('modified_at', models.DateTimeField(auto_now=True, verbose_name='changed at')),
                ('resource_name', models.CharField(editable=False, help_text='Human-readable resource name', max_length=255, verbose_name='resource name')),
                ('extension', models.CharField(editable=False, help_text='Lowercase, without leading dot', max_length=32, verbose_name='extension')),
                ('size', models.PositiveIntegerField(default=0, editable=False, verbose_name='size')),
                ('checksum', models.CharField(editable=False, max_length=64, verbose_name='checksum')),
                ('uploaded_at', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='uploaded at')),
                ('file', paper_uploads.models.fields.base.DynamicStorageFileField(verbose_name='file')),
                ('display_name', models.CharField(blank=True, max_length=255, verbose_name='display name')),
                ('title', models.CharField(blank=True, help_text='The title is being used as a tooltip when the user hovers the mouse over the image', max_length=255, verbose_name='title')),
                ('description', models.TextField(blank=True, help_text='This text will be used by screen readers, search engines, or when the image cannot be loaded', verbose_name='description')),
            ],
            options={
                'verbose_name': 'SVG file',
                'verbose_name_plural': 'SVG files',
                'abstract': False,
                'default_permissions': (),
            },
            bases=(paper_uploads.models.mixins.EditableResourceMixin, paper_uploads.models.mixins.FileFieldProxyMixin, paper_uploads.models.mixins.FileProxyMixin, models.Model),
        ),
    ]



import datetime
import posixpath
import warnings
from collections import OrderedDict
from typing import Any, ClassVar, Dict, Iterable, Optional, Type

import magic
from django.apps import apps
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.contrib.staticfiles.finders import find
from django.contrib.staticfiles.storage import staticfiles_storage
from django.core import checks
from django.core.files import File
from django.core.files.storage import Storage
from django.db import models
from django.db.models.base import ModelBase
from django.db.models.fields.files import FieldFile
from django.db.models.functions import Coalesce
from django.template import loader
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from polymorphic.base import PolymorphicModelBase
from polymorphic.models import PolymorphicModel

from .. import exceptions
from ..conf import FILE_ICON_DEFAULT, FILE_ICON_OVERRIDES, IMAGE_ITEM_VARIATIONS, settings
from ..helpers import (
    _get_item_types,
    _set_item_types,
    build_variations,
    iterate_variation_names,
)
from ..storage import default_storage
from ..utils import cached_method
from ..variations import PaperVariation
from .base import (
    FileFieldResource,
    NoPermissionsMetaBase,
    Resource,
    ResourceBaseMeta,
    SVGFileResourceMixin,
    VersatileImageResourceMixin,
)
from .fields import CollectionItem
from .fields.base import DynamicStorageFileField
from .fields.collection import ContentItemRelation
from .image import VariationalFileField
from .mixins import BacklinkModelMixin, EditableResourceMixin

__all__ = [
    "CollectionItemBase",
    "CollectionFileItemBase",
    "CollectionBase",
    "FilePreviewMixin",
    "FileItemBase",
    "SVGItemBase",
    "ImageItemBase",
    "FileItem",
    "SVGItem",
    "ImageItem",
    "Collection",
    "ImageCollection",
]


class ItemTypesDescriptor:
    """
    Дескриптор, добавляемый к моделям коллекций.
    Возвращает упорядоченый словарь типов элементов коллекции,
    учитывая наследование.
    """

    def __init__(self, name):
        self.name = name

    def __get__(self, instance, cls=None) -> Dict[str, CollectionItem]:
        if cls is None:
            cls = type(instance)

        cache_attname = "_{}__item_type_cache".format(cls.__name__)
        if hasattr(cls, cache_attname):
            return OrderedDict(getattr(cls, cache_attname))

        item_types = OrderedDict()
        parents = [base for base in cls.__mro__ if issubclass(base, CollectionBase)]
        for base in reversed(parents):
            base_item_types = _get_item_types(base)
            if base_item_types is not None:
                undefined_item_types = set(item_types.keys()).difference(
                    base_item_types
                )
                item_types.update(base_item_types)

                # delete overridden item types
                for item_type_name in undefined_item_types:
                    if hasattr(base, item_type_name):
                        del item_types[item_type_name]

        # cache result in class object
        setattr(cls, cache_attname, tuple(item_types.items()))
        return item_types


class CollectionMeta(NoPermissionsMetaBase, ModelBase):
    """
    Хак, при котором вместо наследования создаются прокси-модели,
    если явно не указано обратное.
    """

    @classmethod
    def __prepare__(cls, name, bases):
        return OrderedDict()

    def __new__(mcs, name, bases, attrs, **kwargs):  # noqa: N804
        # set proxy=True by default
        meta = attrs.pop("Meta", None)
        if meta is None:
            meta = type("Meta", (), {"proxy": True})
        else:
            meta_attrs = meta.__dict__.copy()
            meta_attrs.setdefault("proxy", True)
            meta = type("Meta", meta.__bases__, meta_attrs)
        attrs["Meta"] = meta

        # сохраняем явно объявленные (не унаследованные) типы элементов коллекции
        # в приватное поле класса
        item_types = OrderedDict()
        for key, value in list(attrs.items()):
            if isinstance(value, CollectionItem):
                item_types[key] = value

        new_class = super().__new__(mcs, name, bases, attrs, **kwargs)
        _set_item_types(new_class, item_types)
        new_class.item_types = ItemTypesDescriptor("item_types")
        return new_class


class CollectionManager(models.Manager):
    """
    Из-за того, что в большинстве случаев коллекции являются прокси-моделями,
    любые запросы от имени такой модели будут работать с экземплярами конкретной
    модели (обычно Collection).

    Например, запрос
        MyProxyCollection.objects.all()
    вернёт все коллекции, а не только экземпляры MyCustomCollection.

    Данный менеджер ограничивает область действия подобных запросов.
    Для прокси-моделей запросы работают только с экземплярами данной прокси-модели.
    Для конкретных моделей поведение стандартное: запрос вернёт все экземпляры таблицы.
    """

    def get_queryset(self):
        if self.model._meta.proxy:
            collection_ct = ContentType.objects.get_for_model(
                self.model,
                for_concrete_model=False
            )
            return super().get_queryset().filter(
                collection_content_type=collection_ct
            )
        else:
            return super().get_queryset()


class CollectionBase(BacklinkModelMixin, metaclass=CollectionMeta):
    collection_content_type = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE,
        editable=False,
        related_name="+"
    )
    concrete_collection_content_type = models.ForeignKey(
        ContentType,
        null=True,
        on_delete=models.CASCADE,
        editable=False,
        related_name="+"
    )
    items = ContentItemRelation(  # TODO: deprecated since v0.10.1
        "paper_uploads.CollectionItemBase",
        content_type_field="collection_content_type",
        object_id_field="collection_id",
        for_concrete_model=False,
    )
    created_at = models.DateTimeField(
        _("created at"),
        default=now,
        editable=False
    )
    modified_at = models.DateTimeField(
        _("changed at"),
        default=now,
        editable=False
    )

    default_mgr = models.Manager()  # fix migrations manager
    objects = CollectionManager()

    class Meta:
        proxy = False
        abstract = True
        default_permissions = ()
        default_manager_name = "default_mgr"
        verbose_name = _("collection")
        verbose_name_plural = _("collections")

    def __iter__(self):
        return self.get_items().iterator()

    def save(self, *args, **kwargs):
        if self.collection_content_type_id is None:
            self.collection_content_type = ContentType.objects.get_for_model(
                self, for_concrete_model=False
            )
        if self.concrete_collection_content_type_id is None:
            self.concrete_collection_content_type = ContentType.objects.get_for_model(
                self, for_concrete_model=True
            )
        super().save(*args, **kwargs)

    def delete(self, using=None, keep_parents=False):
        # Удаляем элементы группами, из-за рекурсии.
        # Удалить элементы с помощью `.all().delete()` - нельзя. Из-за того,
        # что в файле (django/db/models/deletion.py:284) указано следующее:
        #   model = new_objs[0].__class__
        # Это приводит к тому, что все полиморфные модели удаляются через модель
        # первого экземпляра в QuerySet.
        for item_type in self.item_types:
            self.get_items(item_type).delete()

        super().delete(using=using, keep_parents=keep_parents)

    @classmethod
    def _check_fields(cls, **kwargs):
        errors = super()._check_fields(**kwargs)
        for field in cls.item_types.values():
            errors.extend(field.check(**kwargs))
        return errors

    @classmethod
    def get_item_model(cls, item_type: str) -> 'Type[CollectionItemBase]':
        if item_type not in cls.item_types:
            raise exceptions.InvalidItemType(item_type)
        return cls.item_types[item_type].model

    def get_items(self, item_type: str = None) -> 'models.QuerySet[CollectionItemBase]':
        # Получение элементов коллекции работает как через прокси-модель,
        # так и через соответствующую конкретную модель.
        concrete_model_ct = ContentType.objects.get_for_model(self._meta.concrete_model)
        qs = CollectionItemBase.objects.filter(
            concrete_collection_content_type=concrete_model_ct,
            collection_id=self.pk
        ).order_by("order")

        if item_type is None:
            return qs

        if item_type not in self.item_types:
            raise exceptions.InvalidItemType(item_type)

        return qs.filter(type=item_type)

    def get_last_modified(self) -> datetime.datetime:
        """
        Получение даты модификации коллекции с учётом её элементов.
        """
        dates = [self.created_at, self.modified_at]
        item_date = self.get_items().aggregate(
            date=models.Max("modified_at")
        )["date"]

        if item_date is not None:
            dates.append(item_date)

        return max(*dates)


class Collection(CollectionBase):
    VARIATIONS: ClassVar[dict]

    class Meta:
        proxy = False  # явно указываем, что это не прокси-модель


# ======================================================================================


class CollectionItemMetaBase(PolymorphicModelBase, ResourceBaseMeta):
    pass


class CollectionItemBase(EditableResourceMixin, PolymorphicModel, Resource, metaclass=CollectionItemMetaBase):
    """
    Базовый класс элемента коллекции.
    """

    # путь к шаблону, представляющему элемент коллекции в админке
    template_name: Optional[str] = None

    # путь к шаблону, представляющему картинку-превью элемента коллекции в админке
    preview_template_name: Optional[str] = None

    collection_content_type = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE,
        related_name="+"
    )
    concrete_collection_content_type = models.ForeignKey(
        ContentType,
        null=True,
        on_delete=models.CASCADE,
        editable=False,
        related_name="+"
    )
    collection_id = models.IntegerField()
    collection = GenericForeignKey(
        ct_field="collection_content_type",
        fk_field="collection_id",
        for_concrete_model=False,
    )

    type = models.CharField(
        _("type"),
        max_length=32,
        db_index=True,
        editable=False
    )
    order = models.IntegerField(
        _("order"),
        default=0,
        editable=False
    )

    class Meta:
        # Используется обратный порядок полей в составном индексе,
        # т.к. селективность поля collection_id выше, а для поля
        # `collection_content_type` уже есть отдельный индекс.
        index_together = [("collection_id", "collection_content_type")]
        verbose_name = _("item")
        verbose_name_plural = _("items")

    @classmethod
    def check(cls, **kwargs):
        return [
            *super().check(**kwargs),
            *cls._check_template_name(),
        ]

    @classmethod
    def _check_template_name(cls, **kwargs):
        if cls._meta.abstract or cls is CollectionItemBase:
            return []

        errors = []
        if cls.template_name is None:
            errors.append(
                checks.Error(
                    "{} requires a definition of 'template_name'".format(cls.__name__),
                    obj=cls,
                )
            )
        return errors

    @property
    def item_type(self):
        warnings.warn(
            "'item_type' is deprecated in favor of 'type'",
            DeprecationWarning,
            stacklevel=2
        )
        return self.type

    @item_type.setter
    def item_type(self, value: str):
        warnings.warn(
            "'item_type' is deprecated in favor of 'type'",
            DeprecationWarning,
            stacklevel=2
        )
        self.type = value

    def get_next_order_value(self):
        max_order = CollectionItemBase.objects.filter(
            collection_content_type_id=self.collection_content_type_id,
            collection_id=self.collection_id,
        ).aggregate(
            max_order=models.Max(Coalesce("order", 0))
        )["max_order"]

        return 0 if max_order is None else max_order + 1

    def save(self, *args, **kwargs):
        if self.concrete_collection_content_type_id is None and self.collection_content_type_id is not None:
            try:
                collection_cls = self.get_collection_class()
            except exceptions.CollectionModelNotFoundError:
                # TODO: Добавленный в такую коллекцию элемент не сможет быть
                #       получен через get_items().
                pass
            else:
                self.concrete_collection_content_type = ContentType.objects.get_for_model(
                    collection_cls,
                    for_concrete_model=True
                )

        if not self.pk and self.collection_id:
            # Попытка решить проблему того, что при создании коллекции элементы
            # отсортированы в порядке загрузки, а не в порядке добавления.
            # Код ниже не решает проблему полностью, но уменьшает её влияние.
            if self.order is None or self.collection.get_items().filter(order=self.order).exists():
                self.order = self.get_next_order_value()

        super().save(*args, **kwargs)

    @classmethod
    def accept(cls, *args, **kwargs) -> bool:
        """
        Возвращает True, если переданные данные могут быть представлены
        текущей моделью элемента коллекции.
        """
        raise NotImplementedError

    def as_dict(self) -> Dict[str, Any]:
        return {
            **super().as_dict(),
            "collectionId": self.collection_id,
            "itemType": self.type,  # TODO: deprecated
            "type": self.type,
            "order": self.order,
            "preview": self.render_preview(),
        }

    def get_collection_class(self) -> Type[CollectionBase]:
        # Прямое обращение к полю `self.collection_content_type` дёргает БД.
        # Вместо него используется метод `get_for_id()` класса ContentType,
        # который использует общий кэш.
        # (!) Если класс модели для указанного ContentType удалён, то
        # метод вернёт None.
        collection_ct = ContentType.objects.get_for_id(self.collection_content_type_id)

        try:
            return apps.get_model(collection_ct.app_label, collection_ct.model)
        except LookupError:
            raise exceptions.CollectionModelNotFoundError(
                "{}.{}".format(collection_ct.app_label, collection_ct.model)
            )

    def get_item_type_field(self) -> CollectionItem:
        """
        Получение поля CollectionItem, с которым связан текущий элемент.
        """
        collection_cls = self.get_collection_class()

        field = collection_cls.item_types.get(self.type)
        if field is None:
            raise exceptions.CollectionItemNotFoundError()

        # support proxy models
        if self._meta.concrete_model is field.model._meta.concrete_model:
            return field

        raise exceptions.CollectionItemNotFoundError()

    def get_itemtype_field(self) -> Optional[CollectionItem]:
        warnings.warn(
            "'get_itemtype_field' is deprecated in favor of 'get_item_type_field'",
            DeprecationWarning,
            stacklevel=2
        )
        return self.get_item_type_field()

    def attach_to(self, collection: CollectionBase):
        """
        Подключение элемента к коллекции.
        """
        self.collection_content_type = ContentType.objects.get_for_model(
            collection, for_concrete_model=False
        )
        self.concrete_collection_content_type = ContentType.objects.get_for_model(
            collection, for_concrete_model=True
        )
        self.collection_id = collection.pk

        for name, field in collection.item_types.items():
            # support proxy models
            if self._meta.concrete_model is field.model._meta.concrete_model:
                self.type = name
                break
        else:
            raise exceptions.UnsupportedCollectionItemError(type(self).__name__)

    def render_preview(self):
        """ Отображение элемента коллекции в админке """
        context = self.get_preview_context()
        return loader.render_to_string(self.preview_template_name, context)

    def get_preview_context(self):
        return {
            "item": self,
            "width": settings.COLLECTION_ITEM_PREVIEW_WIDTH,
            "height": settings.COLLECTION_ITEM_PREVIEW_HEIGHT,
        }


class CollectionFileItemBase(CollectionItemBase, FileFieldResource):
    """
    Базовый класс элемента галереи, содержащего файл.
    """

    class Meta:
        abstract = True

    @classmethod
    def accept(cls, file: File) -> bool:
        raise NotImplementedError

    def get_file_storage(self) -> Storage:
        try:
            item_type_field = self.get_item_type_field()
        except (exceptions.CollectionModelNotFoundError, exceptions.CollectionItemNotFoundError):
            return default_storage

        storage = item_type_field.options.get("storage") or default_storage
        if callable(storage):
            storage = storage()
        return storage

    @classmethod
    def file_supported(cls, file: File) -> bool:
        warnings.warn(
            "file_supported() is deprecated in favor of accept()",
            DeprecationWarning,
            stacklevel=2
        )
        return cls.accept(file)


class FilePreviewMixin(models.Model):
    """
    Миксина элемента коллекции, добавляющая иконку для файла в админке
    """

    class Meta:
        abstract = True

    def get_preview_url(self):
        extension = self.extension.lower()  # noqa
        extension = FILE_ICON_OVERRIDES.get(extension, extension)
        icon_path_template = "paper_uploads/dist/assets/{}.svg"
        icon_path = icon_path_template.format(extension)
        if find(icon_path) is None:
            icon_path = icon_path_template.format(FILE_ICON_DEFAULT)
        return staticfiles_storage.url(icon_path)

    def get_preview_context(self):
        context = super().get_preview_context()  # noqa
        context.update(preview_url=self.get_preview_url())
        return context


# ======================================================================================


class FileItemBase(FilePreviewMixin, CollectionFileItemBase):
    change_form_class = "paper_uploads.forms.dialogs.collection.ChangeFileItemDialog"
    template_name = "paper_uploads/items/file.html"
    preview_template_name = "paper_uploads/items/preview/file.html"

    file = DynamicStorageFileField(
        _("file"),
        max_length=255,
    )
    display_name = models.CharField(
        _("display name"),
        max_length=255,
        blank=True
    )

    class Meta(CollectionItemBase.Meta):
        abstract = True
        verbose_name = _("File item")
        verbose_name_plural = _("File items")

    def save(self, *args, **kwargs):
        if not self.pk and not self.display_name:
            self.display_name = self.basename
        super().save(*args, **kwargs)

    def get_file_folder(self) -> str:
        try:
            item_type_field = self.get_item_type_field()
        except (exceptions.CollectionModelNotFoundError, exceptions.CollectionItemNotFoundError):
            return settings.COLLECTION_FILES_UPLOAD_TO

        return item_type_field.options.get("upload_to") or settings.COLLECTION_FILES_UPLOAD_TO

    def get_file(self) -> FieldFile:
        return self.file

    @classmethod
    def get_file_field(cls) -> models.FileField:
        return cls._meta.get_field("file")

    def get_caption(self):
        name = self.display_name or self.basename
        if self.extension:
            return "{}.{}".format(name, self.extension)
        return name

    @classmethod
    def accept(cls, file: File) -> bool:
        return True


class SVGItemBase(SVGFileResourceMixin, CollectionFileItemBase):
    change_form_class = "paper_uploads.forms.dialogs.collection.ChangeSVGItemDialog"
    template_name = "paper_uploads/items/svg.html"
    preview_template_name = "paper_uploads/items/preview/svg.html"

    file = DynamicStorageFileField(
        _("file"),
        max_length=255,
    )
    display_name = models.CharField(
        _("display name"),
        max_length=255,
        blank=True
    )

    class Meta(CollectionItemBase.Meta):
        abstract = True
        verbose_name = _("SVG item")
        verbose_name_plural = _("SVG items")

    def save(self, *args, **kwargs):
        if not self.pk and not self.display_name:
            self.display_name = self.basename
        super().save(*args, **kwargs)

    def get_file_folder(self) -> str:
        try:
            item_type_field = self.get_item_type_field()
        except (exceptions.CollectionModelNotFoundError, exceptions.CollectionItemNotFoundError):
            return settings.COLLECTION_FILES_UPLOAD_TO

        return item_type_field.options.get("upload_to") or settings.COLLECTION_FILES_UPLOAD_TO

    def get_file(self) -> FieldFile:
        return self.file

    @classmethod
    def get_file_field(cls) -> models.FileField:
        return cls._meta.get_field("file")

    def get_caption(self):
        name = self.display_name or self.basename
        if self.extension:
            return "{}.{}".format(name, self.extension)
        return name

    @classmethod
    def accept(cls, file: File) -> bool:
        filename, ext = posixpath.splitext(file.name)
        return ext.lower() == ".svg"


class ImageItemBase(VersatileImageResourceMixin, CollectionFileItemBase):
    PREVIEW_VARIATIONS = IMAGE_ITEM_VARIATIONS
    change_form_class = "paper_uploads.forms.dialogs.collection.ChangeImageItemDialog"
    template_name = "paper_uploads/items/image.html"
    preview_template_name = "paper_uploads/items/preview/image.html"

    file = VariationalFileField(
        _("file"),
        max_length=255,
    )

    class Meta(CollectionItemBase.Meta):
        abstract = True
        verbose_name = _("Image item")
        verbose_name_plural = _("Image items")

    def get_file_folder(self) -> str:
        try:
            item_type_field = self.get_item_type_field()
        except (exceptions.CollectionModelNotFoundError, exceptions.CollectionItemNotFoundError):
            return settings.COLLECTION_IMAGES_UPLOAD_TO

        return item_type_field.options.get("upload_to") or settings.COLLECTION_IMAGES_UPLOAD_TO

    def get_file(self) -> FieldFile:
        return self.file

    @classmethod
    def get_file_field(cls) -> VariationalFileField:
        return cls._meta.get_field("file")

    def recut_async(self, names: Iterable[str] = ()):
        """
        Превью для админки режутся сразу, а остальное — потом.
        """
        preview_variations = tuple(iterate_variation_names(self.PREVIEW_VARIATIONS))
        self.recut(names=preview_variations)

        if not names:
            names = self.get_variations().keys()

        rest_variations = tuple(set(names).difference(preview_variations))
        super().recut_async(rest_variations)

    @cached_method("_variations_cache")
    def get_variations(self) -> Dict[str, PaperVariation]:
        """
        Перебираем возможные места вероятного определения вариаций и берем
        первое непустое значение. Порядок проверки:
            1) параметр `variations` псевдо-поля `CollectionItem`
            2) член класса галереи VARIATIONS
        К найденному словарю примешиваются вариации для админки.
        """
        if not self.collection_content_type_id:
            raise cached_method.Bypass({})

        collection_cls = self.get_collection_class()
        item_type_field = self.get_item_type_field()
        variation_config = self.get_variation_config(collection_cls, item_type_field)
        return build_variations(variation_config)

    @classmethod
    def accept(cls, file: File) -> bool:
        mimetype = magic.from_buffer(file.read(1024), mime=True)
        file.seek(0)  # reset file position after mimetype detection
        basetype, subtype = mimetype.split("/", 1)
        return basetype == "image"

    @classmethod
    def get_variation_config(
        cls, collection_cls: Type[CollectionBase], item_type_field: CollectionItem,
    ) -> Dict[str, Any]:
        if "variations" in item_type_field.options:
            variations = item_type_field.options["variations"]
        else:
            variations = getattr(collection_cls, "VARIATIONS", None)

        variations = variations or {}
        variations = dict(cls.PREVIEW_VARIATIONS, **variations)
        return variations


class FileItem(FileItemBase):
    pass


class SVGItem(SVGItemBase):
    pass


class ImageItem(ImageItemBase):
    pass


# ==============================================================================


class ImageCollection(Collection):
    """
    Коллекция, позволяющая хранить только изображения.
    """
    image = CollectionItem(ImageItem)

    @classmethod
    def get_configuration(cls) -> Dict[str, Any]:
        return {
            "strictImageValidation": True,
            "acceptFiles": [
                "image/bmp",
                "image/gif",
                "image/jpeg",
                "image/png",
                # "image/svg+xml",
                "image/tiff",
                "image/webp",
            ],
        }



from typing import Dict, Type

from django.db import models
from django.db.models.fields import Field

from ..helpers import build_variations
from ..models.base import VersatileImageResourceMixin
from ..models.collection import CollectionBase
from ..models.fields.collection import CollectionItem
from ..variations import PaperVariation


def get_model_name(model: Type[models.Model]) -> str:
    return "{}.{}".format(
        model._meta.app_label,
        model.__name__
    )


def is_variations_allowed(model: Type[models.Model]) -> bool:
    """
    Возвращает True, если модель представляет из себя класс изображения с вариациями.
    """
    return issubclass(model, VersatileImageResourceMixin)


def is_collection(model: Type[models.Model]) -> bool:
    """
    Возвращает True, если model - коллекция.
    """
    return issubclass(model, CollectionBase)


def get_field_variations(field: Field) -> Dict[str, PaperVariation]:
    """
    Получение списка вариаций для поля.
    """
    variation_config = getattr(field, "variations", {}).copy()
    return build_variations(variation_config)


def get_collection_variations(
    collection_cls: Type[CollectionBase], item_type_field: CollectionItem
) -> Dict[str, PaperVariation]:
    """
    Получение списка вариаций для коллекции.
    """
    image_model = item_type_field.model
    variation_config = image_model.get_variation_config(collection_cls, item_type_field)
    return build_variations(variation_config)



import json
from django.template import library
from django.core.serializers.json import DjangoJSONEncoder
from django.utils.html import _json_script_escapes
from django.utils.safestring import mark_safe

register = library.Library()


@register.filter
@register.filter(name="json", is_safe=True)
def do_json(value):
    data = json.dumps(value, cls=DjangoJSONEncoder).translate(_json_script_escapes)
    return mark_safe(data)



# Generated by Django 4.0.1 on 2022-02-17 14:27

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('paper_uploads', '0004_remove_fileitem_created_at_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='collection',
            name='modified_at',
            field=models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='changed at'),
        ),
    ]



# Generated by Django 4.0.1 on 2022-02-16 10:39

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('paper_uploads', '0001_squashed_0009_alter_collectionitembase_polymorphic_ctype_and_more'),
    ]

    operations = [
        migrations.RenameField(
            model_name='fileitem',
            old_name='basename',
            new_name='resource_name',
        ),
        migrations.RenameField(
            model_name='imageitem',
            old_name='basename',
            new_name='resource_name',
        ),
        migrations.RenameField(
            model_name='svgitem',
            old_name='basename',
            new_name='resource_name',
        ),
        migrations.RenameField(
            model_name='uploadedfile',
            old_name='basename',
            new_name='resource_name',
        ),
        migrations.RenameField(
            model_name='uploadedimage',
            old_name='basename',
            new_name='resource_name',
        ),
    ]



from django.utils.translation import gettext_lazy as _

from ...typing import Limitations
from ...utils import filesizeformat
from .base import FileResourceWidgetBase


class DisplayFileLimitationsMixin:
    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)  # noqa: F821
        context.update(
            {
                "limitations": self.get_limitations(),
            }
        )
        return context

    def get_limitations(self) -> Limitations:
        """
        Список ограничений, накладываемых на загружаемые файлы.
        Используется только для вывода в виде текста.
        """
        if not isinstance(self, FileResourceWidgetBase):
            return []

        limits = []  # type: Limitations
        configuration = self.get_configuration()
        if not configuration:
            return limits

        if "acceptFiles" in configuration:
            accept_files = configuration["acceptFiles"]
            limits.append(
                (
                    _("Allowed types"),
                    accept_files
                    if isinstance(accept_files, str)
                    else ", ".join(accept_files),
                )
            )
        if "allowedExtensions" in configuration:
            limits.append(
                (_("Allowed extensions"), ", ".join(configuration["allowedExtensions"]))
            )
        if "sizeLimit" in configuration:
            limits.append(
                (_("Maximum file size"), filesizeformat(configuration["sizeLimit"]))
            )

        min_width = configuration.get("minImageWidth", 0)
        min_height = configuration.get("minImageHeight", 0)
        if min_width:
            if min_height:
                limits.append(
                    (
                        _("Minimum dimensions"),
                        _("%(min_width)sx%(min_height)s pixels") % {
                            "min_width": min_width,
                            "min_height": min_height
                        },
                    )
                )
            else:
                limits.append((_("Minimum image width"), _("%s pixels") % min_width))
        elif min_height:
            limits.append((_("Minimum image height"), _("%s pixels") % min_height))

        max_width = configuration.get("maxImageWidth", 0)
        max_height = configuration.get("maxImageHeight", 0)
        if max_width:
            if max_height:
                limits.append(
                    (
                        _("Maximum dimensions"),
                        _("%(max_width)sx%(max_height)s pixels") % {
                            "max_width": max_width,
                            "max_height": max_height
                        },
                    )
                )
            else:
                limits.append((_("Maximum image width"), _("%s pixels") % max_width))
        elif max_height:
            limits.append((_("Maximum image height"), _("%s pixels") % max_height))

        return limits



import io
import os
import tempfile

from django.core import checks
from django.core.exceptions import SuspiciousFileOperation
from django.core.files import File
from django.utils.crypto import get_random_string
from filelock import FileLock, Timeout

from ... import forms
from ...typing import VariationConfig
from .base import DynamicStorageFileField, FileResourceFieldBase


class VariationalFileField(DynamicStorageFileField):
    """
    Из-за того, что вариация может самостоятельно установить свой формат,
    возможна ситуация, когда вариации одного изображения перезапишут вариации
    другого. Например, когда загружаются файлы, отличающиеся только расширением.
    Поэтому мы проверяем все имена будущих вариаций на существование, чтобы
    не допустить перезапись.
    """

    @staticmethod
    def _variations_collapsed(instance, name, storage):
        if storage.exists(name):
            return True

        for variation in instance.get_variations().values():
            variation_filename = variation.get_output_filename(name)
            if storage.exists(variation_filename):
                return True
        return False

    def _find_available_name(self, instance, name, storage):
        max_length = self.max_length
        dir_name, file_name = os.path.split(name)
        file_root, file_ext = os.path.splitext(file_name)
        while self._variations_collapsed(instance, name, storage) or (
            max_length and len(name) > max_length
        ):
            name = os.path.join(
                dir_name, "%s_%s%s" % (file_root, get_random_string(7), file_ext)
            )
            if max_length is None:
                continue
            # Truncate file_root if max_length exceeded.
            truncation = len(name) - max_length
            if truncation > 0:
                file_root = file_root[:-truncation]
                if not file_root:
                    raise SuspiciousFileOperation(
                        "Storage cannot find an available filename for '%s'. "
                        "Please make sure the corresponding file field "
                        "allows sufficient 'max_length'." % name
                    )
                name = os.path.join(
                    dir_name, "%s_%s%s" % (file_root, get_random_string(7), file_ext)
                )
        return name

    def _create_placeholder_files(self, instance, name, storage):
        """
        Создаем файлы-заглушки, которые "забронируют" имена
        итоговых файлов вариаций.

        Разрешает ситуацию, когда два или более файлов с одинаковым
        именем, но разными расширением загружаются одновременно
        в разных процессах / потоках.
        """
        for variation in instance.get_variations().values():
            variation_filename = variation.get_output_filename(name)
            storage.save(variation_filename, File(io.BytesIO(b"dummy")))

    def generate_filename(self, instance, filename):
        name = super().generate_filename(instance, filename)
        storage = instance.get_file_storage()
        lock = FileLock(os.path.join(tempfile.gettempdir(), "paper_uploads.lock"))
        try:
            with lock.acquire(timeout=5):
                available_name = self._find_available_name(instance, name, storage)
                self._create_placeholder_files(instance, available_name, storage)
        except Timeout:
            available_name = self._find_available_name(instance, name, storage)
        return available_name


class ImageField(FileResourceFieldBase):
    def __init__(self, *args, variations: VariationConfig = None, **kwargs):
        kwargs.setdefault("to", "paper_uploads.UploadedImage")
        self.variations = variations or {}
        super().__init__(*args, **kwargs)

    def check(self, **kwargs):
        return [*super().check(**kwargs), *self._check_variations()]

    def _check_variations(self):
        if not self.variations:
            return []
        for name, variation in self.variations.items():
            if name.startswith("_"):
                return [
                    checks.Error(
                        "Variation name can\'t starts with '_': %s" % name, obj=self,
                    )
                ]
        return []

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if "variations" in kwargs:
            del kwargs["variations"]
        return name, path, args, kwargs

    def formfield(self, **kwargs):
        return super().formfield(**{"form_class": forms.ImageField, **kwargs})



from django import forms

from ...models.svg import UploadedSVGFile
from .base import ChangeFileResourceDialogBase


class ChangeUploadedSVGFileDialog(ChangeFileResourceDialogBase):
    class Meta:
        model = UploadedSVGFile
        fields = ["new_name", "display_name", "title", "description"]
        widgets = {
            "description": forms.Textarea(attrs={
                "rows": 2
            })
        }



from datetime import timedelta
from enum import Enum, auto
from typing import Type

from anytree import LevelOrderIter
from django.contrib.contenttypes.models import ContentType
from django.core.management import BaseCommand
from django.db import DEFAULT_DB_ALIAS, transaction
from django.utils.timezone import now

from ... import helpers
from ...models.base import FileResource
from ...models.collection import Collection, CollectionBase, CollectionItemBase
from ...models.mixins import BacklinkModelMixin
from ..prompt import prompt_action


class Step(Enum):
    CHECK_CONTENT_TYPES = auto()
    CHECK_OWNERSHIP = auto()
    CHECK_FILES = auto()
    END = auto()


class ExitException(Exception):
    pass


class Command(BaseCommand):
    help = """
    Находит некорректные экземпляры файловых моделей и предлагает их удалить.
    
    Некорректными считаются экземпляры, у которых утерян загруженный файл,
    а также экземпляры, на которые нет ссылки.
    
    Создание экземпляра файловой модели и загрузка в неё файла - это  две отдельные 
    операции. Между ними может пройти какое-то время, особенно при использовании 
    django-rq. Для того, чтобы сохранить экземпляры, в которые ещё не загрузились файлы,
    эта команда пропускает те экземпляры, которые созданы недавно.  Временной интервал 
    задаётся через параметр `--threshold` и по умолчанию составляет 24 часа.
    """
    verbosity = None
    database = DEFAULT_DB_ALIAS

    _step = Step.CHECK_CONTENT_TYPES
    _check_content_types = False
    _check_ownership = False
    _check_file_existence = False

    def add_arguments(self, parser):
        parser.add_argument(
            "--database",
            action="store",
            dest="database",
            default=DEFAULT_DB_ALIAS,
            help="Nominates the database to use. Defaults to the 'default' database.",
        )
        parser.add_argument(
            "-c", "--check-content-types",
            action="store_true",
            default=False,
            help="Check collection ContentType.",
        )
        parser.add_argument(
            "-o", "--check-ownership",
            action="store_true",
            default=False,
            help="Check `owner_XXX` fields.",
        )
        parser.add_argument(
            "-f", "--check-file-existence",
            action="store_true",
            default=False,
            help="Check file existence.",
        )
        parser.add_argument(
            "--min-age",
            type=int,
            default=24 * 3600,
            help="Minimum instance age in seconds to look for",
        )

    def handle(self, *args, **options):
        self.options = options
        self.verbosity = options["verbosity"]
        self.database = options["database"]

        check_content_types = self.options["check_content_types"]
        check_ownership = self.options["check_ownership"]
        check_file_existence = self.options["check_file_existence"]
        check_all = not any([
            check_content_types,
            check_ownership,
            check_file_existence,
        ])

        self._check_content_types = check_content_types or check_all
        self._check_ownership = check_ownership or check_all
        self._check_file_existence = check_file_existence or check_all

        try:
            self.loop()
        except ExitException:
            return

    def loop(self):
        while True:
            if self._step is Step.CHECK_CONTENT_TYPES:
                self.clean_invalid_content_types()
            elif self._step is Step.CHECK_OWNERSHIP:
                self.clean_invalid_ownership()
            elif self._step is Step.CHECK_FILES:
                self.clean_file_existence()
            else:
                return

    def clean_invalid_content_types(self):
        """
        Удаление коллекций, у которых:
        *) отсутствует модель, заданная через ContentType.
        """
        if not self._check_content_types:
            self._step = Step.CHECK_OWNERSHIP
            return

        for root in helpers.get_collection_trees():
            for node in reversed(tuple(LevelOrderIter(root))):
                model = node.model
                self._clean_invalid_content_types(model)

        self._step = Step.CHECK_OWNERSHIP

    def _clean_invalid_content_types(self, model: Type[Collection]):
        created_before = now() - timedelta(seconds=self.options["min_age"])
        queryset = model.objects.using(self.database).filter(
            created_at__lte=created_before
        )

        objects = set()
        for instance in queryset.iterator():
            content_type_id = instance.collection_content_type_id
            content_type = ContentType.objects.get_for_id(content_type_id)
            model_class = content_type.model_class()
            if model_class is None:
                objects.add(instance)

        if not objects:
            return

        count = len(objects)
        while True:
            action = prompt_action(
                message=(
                    "Found \033[92m%(count)d %(app_label)s.%(model)s\033[0m instance "
                    "that has an invalid content type.\n"
                    "What would you like to do with it?"
                    if count == 1 else
                    "Found \033[92m%(count)d %(app_label)s.%(model)s\033[0m instances "
                    "that have an invalid content type.\n"
                    "What would you like to do with them?"
                ) % {
                    "count": count,
                    "app_label": model._meta.app_label,
                    "model": model.__name__,
                },
                choices=["Skip", "Print", "Delete", "Exit"]
            )

            if action == "Exit":
                raise ExitException

            if action == "Print":
                print("-" * 48)
                for index, obj in enumerate(objects, start=1):
                    content_type_id = obj.collection_content_type_id
                    content_type = ContentType.objects.get_for_id(content_type_id)

                    number = "{})".format(index)
                    print(
                        "{:<3} \033[92m{}.{}\033[0m #{}\n"
                        "    ContentType ID: \033[91m{}\033[0m\n"
                        "    ContentType model: \033[91m{}.{}\033[0m\n"
                        "    Created at: {}".format(
                            number,
                            obj._meta.app_label,
                            obj.__class__.__name__,
                            obj.pk,
                            content_type_id,
                            content_type.app_label,
                            content_type.model,
                            obj.created_at.strftime("%Y-%m-%d %H:%M:%S"),
                        )
                    )
                print("-" * 48)
            elif action == "Delete":
                with transaction.atomic(using=self.database):
                    for obj in objects:
                        content_type_id = obj.collection_content_type_id
                        CollectionItemBase.objects.filter(
                            collection_content_type_id=content_type_id,
                            collection_id=obj.id
                        ).delete()

                        obj.delete()
                break
            else:
                break

    def clean_invalid_ownership(self):
        """
        Удаление ресурсов и коллекций, у которых:
        *) owner_app_label и owner_model_name, ссылаются на несуществующую модель
        *) имя поля, указанное в owner_fieldname отсутствует в классе модели владельца
        """
        if not self._check_ownership:
            self._step = Step.CHECK_FILES
            return

        for root in helpers.get_resource_model_trees():
            for node in reversed(tuple(LevelOrderIter(root))):
                model = node.model
                self._clean_model_ownership(model)

        for root in helpers.get_collection_trees(include_proxy=True):
            for node in reversed(tuple(LevelOrderIter(root))):
                model = node.model

                if model is Collection:
                    continue

                self._clean_model_ownership(model)

        self._step = Step.CHECK_FILES

    def _clean_model_ownership(self, model: Type[BacklinkModelMixin]):
        if not issubclass(model, BacklinkModelMixin):
            return

        query_fields = [
            "pk",
            "owner_app_label",
            "owner_model_name",
            "owner_fieldname"
        ]
        if issubclass(model, CollectionBase):
            query_fields.extend([
                "collection_content_type_id",
            ])

        created_before = now() - timedelta(seconds=self.options["min_age"])
        queryset = model.objects.using(self.database).filter(
            created_at__lte=created_before
        ).only(*query_fields)

        objects = set()
        for instance in queryset.iterator():
            owner_field = instance.get_owner_field()
            if owner_field is None:
                objects.add(instance)

        if not objects:
            return

        count = len(objects)
        while True:
            action = prompt_action(
                message=(
                    "Found \033[92m%(count)d %(app_label)s.%(model)s\033[0m instance "
                    "that has an invalid owner reference.\n"
                    "What would you like to do with it?"
                    if count == 1 else
                    "Found \033[92m%(count)d %(app_label)s.%(model)s\033[0m instances "
                    "that have an invalid owner reference.\n"
                    "What would you like to do with them?"
                ) % {
                    "count": count,
                    "app_label": model._meta.app_label,
                    "model": model.__name__,
                },
                choices=["Skip", "Print", "Delete", "Exit"]
            )

            if action == "Exit":
                raise ExitException

            if action == "Print":
                print("-" * 48)
                for index, obj in enumerate(objects, start=1):
                    number = "{})".format(index)
                    print(
                        "{:<3} \033[92m{}.{}\033[0m #{}\n"
                        "    Owner reference: \033[91m{}.{}\033[0m\n"
                        "    Owner field: \033[91m{}\033[0m\n"
                        "    Created at: {}".format(
                            number,
                            obj._meta.app_label,
                            obj.__class__.__name__,
                            obj.pk,
                            obj.owner_app_label,
                            obj.owner_model_name,
                            obj.owner_fieldname,
                            obj.created_at.strftime("%Y-%m-%d %H:%M:%S"),
                        )
                    )
                print("-" * 48)
            elif action == "Delete":
                with transaction.atomic(using=self.database):
                    # Каждый экземпляр должен быть удалён через свою прокси-модель,
                    # если такая есть. Поэтому удаление через QuerySet не допустимо.
                    for obj in objects:
                        obj.delete()
                break
            else:
                break

    def clean_file_existence(self):
        if not self._check_file_existence:
            self._step = Step.END
            return

        for root in helpers.get_resource_model_trees():
            for node in reversed(tuple(LevelOrderIter(root))):
                model = node.model
                self._clean_file_existence(model)

        self._step = Step.END

    def _clean_file_existence(self, model: Type[FileResource]):
        if not issubclass(model, FileResource):
            return

        created_before = now() - timedelta(seconds=self.options["min_age"])
        queryset = model.objects.using(self.database).filter(
            created_at__lte=created_before
        )

        objects = set()
        for instance in queryset.iterator():
            if not instance.file_exists():
                objects.add(instance)

        if not objects:
            return

        count = len(objects)
        while True:
            action = prompt_action(
                message=(
                    "Found \033[92m%(count)d %(app_label)s.%(model)s\033[0m instance "
                    "that refers to a file that does not exist.\n"
                    "What would you like to do with it?"
                    if count == 1 else
                    "Found \033[92m%(count)d %(app_label)s.%(model)s\033[0m instances "
                    "that refer to a file that does not exist.\n"
                    "What would you like to do with them?"
                ) % {
                    "count": count,
                    "app_label": model._meta.app_label,
                    "model": model.__name__,
                },
                choices=["Skip", "Print", "Delete", "Exit"]
            )

            if action == "Exit":
                raise ExitException

            if action == "Print":
                print("-" * 48)
                for index, obj in enumerate(objects, start=1):
                    number = "{})".format(index)
                    print(
                        "{:<3} \033[92m{}.{}\033[0m #{}\n"
                        "    Created at: {}".format(
                            number,
                            obj._meta.app_label,
                            obj.__class__.__name__,
                            obj.pk,
                            obj.created_at.strftime("%Y-%m-%d %H:%M:%S"),
                        )
                    )
                print("-" * 48)
            elif action == "Delete":
                with transaction.atomic(using=self.database):
                    # Каждый экземпляр должен быть удалён через свою прокси-модель,
                    # если такая есть. Поэтому удаление через QuerySet не допустимо.
                    for obj in objects:
                        obj.delete()
                break
            else:
                break



# Generated by Django 4.0.2 on 2022-03-25 12:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('paper_uploads', '0007_uploadedsvgfile'),
    ]

    operations = [
        migrations.AddField(
            model_name='svgitem',
            name='height',
            field=models.DecimalField(decimal_places=4, default=0, editable=False, max_digits=10, verbose_name='height'),
        ),
        migrations.AddField(
            model_name='svgitem',
            name='width',
            field=models.DecimalField(decimal_places=4, default=0, editable=False, max_digits=10, verbose_name='width'),
        ),
        migrations.AddField(
            model_name='uploadedsvgfile',
            name='height',
            field=models.DecimalField(decimal_places=4, default=0, editable=False, max_digits=10, verbose_name='height'),
        ),
        migrations.AddField(
            model_name='uploadedsvgfile',
            name='width',
            field=models.DecimalField(decimal_places=4, default=0, editable=False, max_digits=10, verbose_name='width'),
        ),
    ]



from typing import Any, Dict

from django.core.files.storage import Storage
from django.db import models
from django.db.models.fields.files import FieldFile
from django.utils.translation import gettext_lazy as _

from ..conf import settings
from ..storage import default_storage
from ..utils import filesizeformat
from .base import FileFieldResource, SVGFileResourceMixin
from .fields.base import DynamicStorageFileField
from .mixins import BacklinkModelMixin, EditableResourceMixin


class UploadedSVGFileBase(SVGFileResourceMixin, BacklinkModelMixin, EditableResourceMixin, FileFieldResource):
    change_form_class = "paper_uploads.forms.dialogs.svg.ChangeUploadedSVGFileDialog"

    file = DynamicStorageFileField(
        _("file"),
        max_length=255,
    )
    display_name = models.CharField(
        _("display name"),
        max_length=255,
        blank=True
    )

    class Meta(FileFieldResource.Meta):
        abstract = True
        verbose_name = _("SVG file")
        verbose_name_plural = _("SVG files")

    def save(self, *args, **kwargs):
        if not self.pk and not self.display_name:
            self.display_name = self.resource_name
        super().save(*args, **kwargs)

    def get_file_folder(self) -> str:
        owner_field = self.get_owner_field()
        return getattr(owner_field, "upload_to", "") or settings.FILES_UPLOAD_TO

    def get_file_storage(self) -> Storage:
        owner_field = self.get_owner_field()
        storage = getattr(owner_field, "storage", None) or default_storage
        if callable(storage):
            storage = storage()
        return storage

    def get_file(self) -> FieldFile:
        return self.file

    @classmethod
    def get_file_field(cls) -> models.FileField:
        return cls._meta.get_field("file")

    def get_caption(self) -> str:
        name = self.display_name or self.resource_name
        if self.extension:
            return "{}.{}".format(name, self.extension)
        return name

    def as_dict(self) -> Dict[str, Any]:
        return {
            **super().as_dict(),
            "name": self.display_name or self.resource_name,
            "file_info": "({ext}, {size})".format(
                ext=self.extension, size=filesizeformat(self.size)
            ),
        }

    @classmethod
    def get_configuration(cls) -> Dict[str, Any]:
        return {
            "acceptFiles": [
                "image/svg+xml",
            ],
        }


class UploadedSVGFile(UploadedSVGFileBase):
    class Meta(UploadedSVGFileBase.Meta):
        pass



from typing import Optional, Type

from django.apps import apps
from django.core import checks
from django.core.exceptions import FieldDoesNotExist
from django.core.files import File
from django.db import models
from django.db.models.utils import make_model_tuple
from django.utils.module_loading import import_string

from ..logging import logger


class BacklinkModelMixin(models.Model):
    """
    Миксина, позволяющая обратиться к полю модели, которое ссылается
    на текущий объект.

    Этот миксин должен использоваться в каждой модели, которая используется
    в полях FileField, ImageField или CollectionField.
    """

    owner_app_label = models.CharField(max_length=100, editable=False)
    owner_model_name = models.CharField(max_length=100, editable=False)
    owner_fieldname = models.CharField(max_length=255, editable=False)

    class Meta:
        abstract = True

    def set_owner_field(self, model: Type[models.Model], field_name: str):
        # check field exists
        field = model._meta.get_field(field_name)

        self.owner_app_label, self.owner_model_name = make_model_tuple(model)
        self.owner_fieldname = field.name

    def get_owner_model(self) -> Optional[Type[models.Model]]:
        if not self.owner_app_label or not self.owner_model_name:
            return

        try:
            return apps.get_model(self.owner_app_label, self.owner_model_name)
        except LookupError:
            logger.debug(
                "Not found model: %s.%s" % (self.owner_app_label, self.owner_model_name)
            )

    def get_owner_field(self) -> Optional[models.Field]:
        owner_model = self.get_owner_model()
        if owner_model is None:
            return

        if not self.owner_fieldname:
            return

        try:
            return owner_model._meta.get_field(self.owner_fieldname)
        except FieldDoesNotExist:
            logger.debug(
                "Not found field '%s' in model %s.%s"
                % (self.owner_app_label, self.owner_model_name, self.owner_fieldname)
            )


class FileProxyMixin:
    """
    Миксин, проксирующий некоторые свойства объекта, возвращаемого методом `get_file()`
    на уровень класса, к которому применен миксин.
    """

    def __init__(self, *args, **kwargs):
        self.__file = None
        super().__init__(*args, **kwargs)

    def __enter__(self):
        return self.open()

    def __exit__(self, exc_type, exc_value, tb):
        self.close()

    def _require_file(self):
        """
        Требование наличия непустой ссылки на файл.
        Физическое существование файла не гарантируется.
        """
        raise NotImplementedError

    def _get_file(self):
        return self.__file

    def _open_file(self, mode):
        self.__file = self.get_file().open(mode)
        return self.__file

    def get_file(self) -> File:
        raise NotImplementedError

    def open(self, mode="rb"):
        """
        Открытие файла.

        Если файл уже открыт, перемещает курсор в начало файла.
        Если файл не допускает перемещение курсора - файл закрывается и открывается
        заново.
        """
        self._require_file()

        file = self._get_file()
        if file is None:
            return self._open_file(mode)

        if file.seekable() and (not hasattr(file.file, "mode") or file.file.mode == mode):
            file.seek(0)
            return file

        # current file is not seekable - reopen it
        file.close()
        return self._open_file(mode)
    open.alters_data = True

    @property
    def closed(self):
        return self.__file is None or self.__file.closed

    def close(self):
        if self.__file is not None:
            self.__file.close()
            self.__file = None


class FileFieldProxyMixin:
    """
    Проксирование некоторых свойств файла (только для чтения) на уровень модели
    """

    @property
    def path(self):
        return self.get_file().path  # noqa: F821

    @property
    def url(self):
        return self.get_file().url  # noqa: F821


class EditableResourceMixin:
    """
    Добавление поля сос ссылкой на класс формы, через которую
    следует редактировать ресурс в интерфейсе администратора.
    """
    change_form_class: Optional[str] = None

    @classmethod
    def check(cls, **kwargs):
        return [
            *super().check(**kwargs),
            *cls._check_form_class(),
        ]

    @classmethod
    def _check_form_class(cls, **kwargs):
        if cls._meta.abstract or cls.change_form_class is None:
            return []

        errors = []

        try:
            import_string(cls.change_form_class)
        except ImportError:
            errors.append(
                checks.Error(
                    "The value of 'change_form_class' refers to '%s', which does "
                    "not exists" % cls.change_form_class,
                    obj=cls,
                )
            )

        return errors



from .collection import CollectionField, CollectionItem
from .file import FileField
from .svg import SVGFileField
from .image import ImageField, VariationalFileField

__all__ = [
    "VariationalFileField",
    "FileField",
    "SVGFileField",
    "ImageField",
    "CollectionField",
    "CollectionItem",
]



# Generated by Django 4.0.2 on 2022-03-25 09:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('paper_uploads', '0005_collection_modified_at'),
    ]

    operations = [
        migrations.AddField(
            model_name='svgitem',
            name='description',
            field=models.TextField(blank=True, help_text='This text will be used by screen readers, search engines, or when the image cannot be loaded', verbose_name='description'),
        ),
        migrations.AddField(
            model_name='svgitem',
            name='title',
            field=models.CharField(blank=True, help_text='The title is being used as a tooltip when the user hovers the mouse over the image', max_length=255, verbose_name='title'),
        ),
    ]



from .base import ResourceAdminBase  # noqa: F401
from .collection import CollectionAdminBase  # noqa: F401
from .file import UploadedFileAdmin  # noqa: F401
from .svg import UploadedSVGFileAdmin  # noqa: F401
from .image import UploadedImageAdmin  # noqa: F401



import os
import time
from functools import lru_cache
from typing import Any, Dict, Generator, Iterable, Iterator, List, Set, Tuple, Type

from anytree import Node
from django.apps import apps
from django.core import exceptions
from django.core.exceptions import ObjectDoesNotExist
from django.db import DEFAULT_DB_ALIAS, models

from .conf import settings
from .logging import logger
from .typing import VariationConfig
from .utils import lowercased_dict_keys
from .variations import PaperVariation

# Перечень допустимых версий вариаций
ALLOWED_VERSIONS = {"webp", "2x", "3x", "4x"}

# Максимальное количество попыток чтения из БД при получении
# экземпляра модели. Сделано потому, что каждый запрос оборачивается
# в транзакцию. Из-за этого вызов `recut()` (даже после метода `save()`)
# не гарантирует, что данные уже попали в БД.
# TODO: поискать альтернативные решения
MAX_DB_READ_ATTEMPTS = 3


def get_filename(filename: str) -> str:
    basename = os.path.basename(filename)
    name, _ = os.path.splitext(basename)
    return name


def get_extension(filename: str) -> str:
    basename = os.path.basename(filename)
    _, extension = os.path.splitext(basename)
    return extension.lstrip(".")


def _variation_name(name: str, scale_factor: int = 1, webp: bool = False) -> str:
    if webp:
        variation_name = "{}_webp".format(name)
    else:
        variation_name = name

    if scale_factor == 1:
        pass
    else:
        variation_name = "{}_{}x".format(variation_name, scale_factor)

    return variation_name


def generate_scaled_versions(
    name: str, config: VariationConfig, scale_factor: int = 1, webp: bool = False
) -> Iterator[PaperVariation]:
    """
    Геренирует Retina-версию вариации с опциональной WebP-версией того же размера.
    """
    scaled_size = tuple(x * scale_factor for x in config.get("size", (0, 0)))

    variation_config = dict(
        config,
        name=_variation_name(name, scale_factor),
        size=scaled_size,
        max_width=scale_factor * config.get("max_width", 0),
        max_height=scale_factor * config.get("max_height", 0),
    )

    yield PaperVariation(**variation_config)

    if webp:
        variation_config = dict(
            config,
            name=_variation_name(name, scale_factor, webp=True),
            size=scaled_size,
            max_width=scale_factor * config.get("max_width", 0),
            max_height=scale_factor * config.get("max_height", 0),
            format="webp"
        )

        yield PaperVariation(**variation_config)


def generate_variation_versions(
    name: str, config: VariationConfig, versions: Set[str],
) -> Iterator[PaperVariation]:
    """
    Геренирует все указанные версии вариации по словарю конфигурации
    """
    webp = "webp" in versions and config.get("format", "").upper() != "WEBP"
    yield from generate_scaled_versions(name, config, scale_factor=1, webp=webp)
    if "2x" in versions:
        yield from generate_scaled_versions(name, config, scale_factor=2, webp=webp)
    if "3x" in versions:
        yield from generate_scaled_versions(name, config, scale_factor=3, webp=webp)
    if "4x" in versions:
        yield from generate_scaled_versions(name, config, scale_factor=4, webp=webp)


def build_variations(options: Dict[str, VariationConfig]) -> Dict[str, PaperVariation]:
    """
    Создание объектов вариаций из словаря конфигураций.
    """
    variations = {}
    for name, config in options.items():
        new_config = lowercased_dict_keys(settings.VARIATION_DEFAULTS or {})
        new_config.update(config)

        versions = set(v.lower() for v in new_config.get("versions", ()))
        unknown_versions = versions.difference(ALLOWED_VERSIONS)
        if unknown_versions:
            raise ValueError(
                "unknown variation versions: {}".format(", ".join(unknown_versions))
            )

        variation_name = new_config.get("name", name)

        all_versions = generate_variation_versions(variation_name, new_config, versions)
        for variation in all_versions:
            if variation_name == variation.name:  # определение корневой вариации
                variations[variation_name] = variation
            else:
                variations.setdefault(variation.name, variation)

    return variations


def iterate_variation_names(options: Dict[str, VariationConfig]) -> Iterator[str]:
    """
    Перечисляет имена вариаций из словаря конфигураций с учётом параметра `versions`.
    """
    def iterate_variation_versions(name: str, scale_factor: int = 1, webp: bool = False) -> Iterator[str]:
        yield _variation_name(name, scale_factor)
        if webp:
            yield _variation_name(name, scale_factor, webp=True)

    for name, config in options.items():
        new_config = lowercased_dict_keys(settings.VARIATION_DEFAULTS or {})
        new_config.update(config)

        versions = set(v.lower() for v in config.get("versions", ()))
        webp = "webp" in versions and config.get("format", "").upper() != "WEBP"

        yield from iterate_variation_versions(name, scale_factor=1, webp=webp)
        if "2x" in versions:
            yield from iterate_variation_versions(name, scale_factor=2, webp=webp)
        if "3x" in versions:
            yield from iterate_variation_versions(name, scale_factor=3, webp=webp)
        if "4x" in versions:
            yield from iterate_variation_versions(name, scale_factor=4, webp=webp)


@lru_cache()
def get_resource_model_trees(include_proxy=False) -> Tuple[Node]:
    """
    Возвращает иерархии классов ресурсов в виде anytree.Node.
    """
    from .models.base import Resource

    resource_models = [
        model
        for model in apps.get_models()
        if issubclass(model, Resource) and (include_proxy is True or model._meta.proxy is False)
    ]

    resource_bases = {
        model: [
            base
            for base in model.__mro__[1:]
            if base in resource_models
        ]
        for model in resource_models
    }

    # Иерархии ресурсов, отсортированные по длине
    ordered_resource_bases = sorted(resource_bases.items(), key=lambda p: len(p[1]), reverse=True)

    # Ссылки на все узлы всех деревьев
    node_map = {}

    # Список корневых узлов деревьев
    trees = []  # type: list[Node]

    def _get_or_create_node(model, parent=None):
        if model in node_map:
            return node_map[model]
        else:
            node = Node(model.__name__, model=model, parent=parent)
            node_map[model] = node

            if parent is None:
                trees.append(node)

            return node

    # Перебор цепочек классов от самых длинных к самым коротким гарантирует,
    # что родительские классы появятся в дереве раньше дочерних. Это
    # позволит использовать упрощенный алгоритм построения дерева.
    for model, bases in ordered_resource_bases:
        base_parent = None
        for base in reversed(bases):
            base_parent = _get_or_create_node(base, base_parent)

        _get_or_create_node(model, base_parent)

    return tuple(trees)


@lru_cache()
def get_collection_trees(include_proxy=False) -> Tuple[Node]:
    """
    Возвращает иерархии классов коллекций в виде anytree.Node.
    """
    from .models.collection import Collection

    collection_models = [
        model
        for model in apps.get_models()
        if issubclass(model, Collection) and (include_proxy is True or model._meta.proxy is False)
    ]

    collection_bases = {
        model: [
            base
            for base in model.__mro__[1:]
            if base in collection_models
        ]
        for model in collection_models
    }

    # Иерархии коллекций, отсортированные по длине
    ordered_collection_bases = sorted(collection_bases.items(), key=lambda p: len(p[1]), reverse=True)

    # Ссылки на все узлы всех деревьев
    node_map = {}

    # Список корневых узлов деревьев
    trees = []  # type: list[Node]

    def _get_or_create_node(model, parent=None):
        if model in node_map:
            return node_map[model]
        else:
            node = Node(model.__name__, model=model, parent=parent)
            node_map[model] = node

            if parent is None:
                trees.append(node)

            return node

    # Перебор цепочек классов от самых длинных к самым коротким гарантирует,
    # что родительские классы появятся в дереве раньше дочерних. Это
    # позволит использовать упрощенный алгоритм построения дерева.
    for model, bases in ordered_collection_bases:
        base_parent = None
        for base in reversed(bases):
            base_parent = _get_or_create_node(base, base_parent)

        _get_or_create_node(model, base_parent)

    return tuple(trees)


def get_instance(
    app_label: str, model_name: str, object_id: int, using: str = DEFAULT_DB_ALIAS
):
    """
    Получение экземпляра модели по названию приложения, модели и ID.
    """
    model_class = apps.get_model(app_label, model_name)
    attempts = 1
    while True:
        try:
            return model_class._base_manager.using(using).get(pk=object_id)
        except ObjectDoesNotExist:
            # delay recheck if transaction not committed yet
            attempts += 1
            if attempts > MAX_DB_READ_ATTEMPTS:
                logger.exception("Instance #%s not found" % object_id)
                raise
            else:
                time.sleep(1)


def run_validators(value: Any, validators: Iterable[Any]):
    """
    Based on `django.forms.fields.run_validators` method.
    """
    errors = []  # type: List[Any]
    for v in validators:
        try:
            v(value)
        except exceptions.ValidationError as e:
            errors.extend(e.error_list)

    if errors:
        raise exceptions.ValidationError(errors)


def iterate_parent_models(model: models.Model) -> Generator[Type[models.Model], Any, None]:
    """
    Итерация модели и её родительских классов-моделей.
    """
    for klass in model.__mro__:
        if issubclass(klass, models.Model):
            yield klass


def _get_item_types(cls):
    """
    Получение явно объявленных (не унаследованных) типов элементов коллекции
    из её класса.
    """
    attname = "_{}__item_types".format(cls.__name__)
    return getattr(cls, attname, None)


def _set_item_types(cls, value):
    attname = "_{}__item_types".format(cls.__name__)
    setattr(cls, attname, value)



import json
from typing import Dict

from django.contrib.contenttypes.models import ContentType
from django.forms import widgets
from django.utils.functional import cached_property


class ResourceWidgetBase(widgets.Widget):
    owner_app_label = None
    owner_model_name = None
    owner_fieldname = None

    @cached_property
    def model(self):
        return self.choices.queryset.model  # noqa: F821

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        context.update(
            {
                "content_type": ContentType.objects.get_for_model(
                    self.model, for_concrete_model=False
                ),
                "owner_app_label": self.owner_app_label,
                "owner_model_name": self.owner_model_name,
                "owner_fieldname": self.owner_fieldname,
                "instance": self.get_instance(value) if value else None,
            }
        )
        return context

    def get_instance(self, value):
        return self.model._base_manager.get(pk=value)


class FileResourceWidgetBase(ResourceWidgetBase):
    configuration = None  # type: Dict

    def __init__(self, *args, **kwargs):
        self.configuration = self.configuration or {}
        super().__init__(*args, **kwargs)

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)  # noqa: F821
        context.update(
            {
                "configuration": json.dumps(self.get_configuration()),
            }
        )
        return context

    def get_configuration(self):
        configuration_method = getattr(self.model, "get_configuration", None)  # noqa: F821
        if configuration_method is not None and callable(configuration_method):
            config = configuration_method()
        else:
            config = {}

        config.update(self.configuration)
        return config



from typing import Any, Dict

import django
from django.core import checks
from django.core.files import File
from django.db import models
from django.db.models.fields.files import FieldFile, FileDescriptor
from django.db.models.signals import post_delete

from ... import validators


class ResourceFieldBase(models.OneToOneField):
    """
    Базовый класс для ссылок на ресурсы.
    """

    def __init__(self, verbose_name=None, **kwargs):
        kwargs.setdefault("null", True)
        kwargs.setdefault("related_name", "+")
        kwargs.setdefault("on_delete", models.SET_NULL)
        kwargs["verbose_name"] = verbose_name
        super().__init__(**kwargs)

    def check(self, **kwargs):
        return [
            *super().check(**kwargs),
            *self._check_relation()
        ]

    def _check_relation(self):
        from ...models.base import Resource

        return self._check_relation_class(
            Resource,
            "Field defines a relation with model '%s', "
            "which is not subclass of Resource model",
        )

    def _check_relation_class(self, base, error_message):
        rel_is_string = isinstance(self.remote_field.model, str)
        if rel_is_string:
            return []

        model_name = (
            self.remote_field.model
            if rel_is_string
            else self.remote_field.model._meta.object_name
        )

        if not issubclass(self.remote_field.model, base):
            return [
                checks.Error(
                    error_message % model_name,
                    obj=self
                )
            ]
        return []

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if "null" in kwargs and self.null:
            del kwargs["null"]
        if "related_name" in kwargs and kwargs["related_name"] == "+":
            del kwargs["related_name"]
        return name, path, args, kwargs

    def formfield(self, **kwargs):
        return super().formfield(
            **{
                "owner_app_label": self.opts.app_label.lower(),
                "owner_model_name": self.opts.model_name.lower(),
                "owner_fieldname": self.name,
                **kwargs,
            }
        )

    def run_validators(self, value):
        """
        Валидаторы мы хотим запускать по отношению в экземпляру File,
        а не к ID экземпляа модели.
        """
        pass

    def contribute_to_class(self, cls, *args, **kwargs):
        super().contribute_to_class(cls, *args, **kwargs)
        if not cls._meta.abstract:
            post_delete.connect(self.post_delete_callback, sender=cls)

    def post_delete_callback(self, **kwargs):
        """
        При удалении модели-владельца поля, удаляем и связанный
        в ним экземпляр ресурса.
        """
        owner_instance = kwargs["instance"]
        resource_id = getattr(owner_instance, self.attname)
        if resource_id:
            # Удаление через `owner_instance.file.delete()` приведёт к зацикленности
            # в случае использования `on_delete=CASCADE`. Поэтому используем фильтрацию
            # по pk.
            self.related_model._base_manager.filter(pk=resource_id).delete()


class FileResourceFieldBase(ResourceFieldBase):
    """
    Базовый класс для полей, которые загружают файлы.
    """
    def __init__(self, *args, upload_to="", storage=None, **kwargs):
        self.storage = storage
        self.upload_to = upload_to
        super().__init__(*args, **kwargs)

    def check(self, **kwargs):
        return [
            *super().check(**kwargs),
            *self._check_upload_to(),
        ]

    def _check_upload_to(self):
        if callable(self.upload_to):
            return [
                checks.Error(
                    "%s's 'upload_to' argument must be a string, not a callable." % self.__class__.__name__,
                    obj=self,
                )
            ]
        elif isinstance(self.upload_to, str) and self.upload_to.startswith('/'):
            return [
                checks.Error(
                    "%s's 'upload_to' argument must be a relative path, not an "
                    "absolute path." % self.__class__.__name__,
                    obj=self,
                    id="fields.E202",
                    hint="Remove the leading slash.",
                )
            ]
        else:
            return []

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs["storage"] = self.storage
        kwargs["upload_to"] = self.upload_to
        return name, path, args, kwargs

    def formfield(self, **kwargs):
        return super().formfield(**{"configuration": self.get_configuration(), **kwargs})

    def get_configuration(self) -> Dict[str, Any]:
        """
        Превращает Django-валидаторы в словарь конфигурации,
        который может использоваться для вывода или проверки
        на стороне клиента.
        """
        config = {}
        for v in self.validators:
            if isinstance(v, validators.MimeTypeValidator):
                config["acceptFiles"] = v.allowed
            elif isinstance(v, validators.ExtensionValidator):
                config["allowedExtensions"] = v.allowed
            elif isinstance(v, validators.MaxSizeValidator):
                config["sizeLimit"] = v.limit_value
            elif isinstance(v, validators.ImageMinSizeValidator):
                config["minImageWidth"] = v.width_limit
                config["minImageHeight"] = v.height_limit
            elif isinstance(v, validators.ImageMaxSizeValidator):
                config["maxImageWidth"] = v.width_limit
                config["maxImageHeight"] = v.height_limit
        return config


class DynamicStorageFieldFile(FieldFile):
    def __init__(self, instance, field, name):
        super(FieldFile, self).__init__(None, name)
        self.instance = instance
        self.field = field
        self.storage = instance.get_file_storage()
        self._committed = True

    def __getstate__(self):
        state = super().__getstate__()
        if django.VERSION <= (3, 0):
            state.setdefault("field", self.field)
            state.setdefault("instance", self.instance)

        state.setdefault("storage", self.storage)
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        # Попытка восстановления storage в __setstate__ терпит неудачу из-за
        # того, что в instance ещё нет полей owner_app_label и прочих.


class DynamicStorageFileDescriptor(FileDescriptor):
    def __get__(self, instance, cls=None):
        if instance is None:
            return self

        if django.VERSION <= (3, 1):
            # The instance dict contains whatever was originally assigned
            # in __set__.
            if self.field.name in instance.__dict__:
                file = instance.__dict__[self.field.name]
            else:
                instance.refresh_from_db(fields=[self.field.name])
                file = getattr(instance, self.field.name)

            # If this value is a string (instance.file = "path/to/file") or None
            # then we simply wrap it with the appropriate attribute class according
            # to the file field. [This is FieldFile for FileFields and
            # ImageFieldFile for ImageFields; it's also conceivable that user
            # subclasses might also want to subclass the attribute class]. This
            # object understands how to convert a path to a file, and also how to
            # handle None.
            if isinstance(file, str) or file is None:
                attr = self.field.attr_class(instance, self.field, file)
                instance.__dict__[self.field.name] = attr

            # Other types of files may be assigned as well, but they need to have
            # the FieldFile interface added to them. Thus, we wrap any other type of
            # File inside a FieldFile (well, the field's attr_class, which is
            # usually FieldFile).
            elif isinstance(file, File) and not isinstance(file, FieldFile):
                file_copy = self.field.attr_class(instance, self.field, file.name)
                file_copy.file = file
                file_copy._committed = False
                instance.__dict__[self.field.name] = file_copy

            # Finally, because of the (some would say boneheaded) way pickle works,
            # the underlying FieldFile might not actually itself have an associated
            # file. So we need to reset the details of the FieldFile in those cases.
            elif isinstance(file, FieldFile) and not hasattr(file, "field"):
                file.instance = instance
                file.field = self.field
                file.storage = instance.get_file_storage()

            # Make sure that the instance is correct.
            elif isinstance(file, FieldFile) and instance is not file.instance:
                file.instance = instance

            # That was fun, wasn't it?
            return instance.__dict__[self.field.name]
        else:
            # The instance dict contains whatever was originally assigned
            # in __set__.
            file = super().__get__(instance, cls)

            # If this value is a string (instance.file = "path/to/file") or None
            # then we simply wrap it with the appropriate attribute class according
            # to the file field. [This is FieldFile for FileFields and
            # ImageFieldFile for ImageFields; it's also conceivable that user
            # subclasses might also want to subclass the attribute class]. This
            # object understands how to convert a path to a file, and also how to
            # handle None.
            if isinstance(file, str) or file is None:
                attr = self.field.attr_class(instance, self.field, file)
                instance.__dict__[self.field.attname] = attr

            # Other types of files may be assigned as well, but they need to have
            # the FieldFile interface added to them. Thus, we wrap any other type of
            # File inside a FieldFile (well, the field's attr_class, which is
            # usually FieldFile).
            elif isinstance(file, File) and not isinstance(file, FieldFile):
                file_copy = self.field.attr_class(instance, self.field, file.name)
                file_copy.file = file
                file_copy._committed = False
                instance.__dict__[self.field.attname] = file_copy

            # Finally, because of the (some would say boneheaded) way pickle works,
            # the underlying FieldFile might not actually itself have an associated
            # file. So we need to reset the details of the FieldFile in those cases.
            elif isinstance(file, FieldFile) and not hasattr(file, "field"):
                file.instance = instance
                file.field = self.field
                file.storage = instance.get_file_storage()

            # Make sure that the instance is correct.
            elif isinstance(file, FieldFile) and instance is not file.instance:
                file.instance = instance

            # That was fun, wasn't it?
            return instance.__dict__[self.field.attname]


class DynamicStorageFileField(models.FileField):
    """
    FileField, который проксирует вызов метода `generate_filename()` на модель,
    в которой определено данное поле.
    """
    attr_class = DynamicStorageFieldFile
    descriptor_class = DynamicStorageFileDescriptor

    def __init__(self, verbose_name=None, name=None, upload_to='', storage=None, **kwargs):
        self._primary_key_set_explicitly = "primary_key" in kwargs
        kwargs.setdefault("max_length", 255)
        super(models.FileField, self).__init__(verbose_name, name, **kwargs)

    def check(self, **kwargs):
        return [
            *super(models.FileField, self).check(**kwargs),
            *self._check_primary_key(),
        ]

    def deconstruct(self):
        name, path, args, kwargs = super(models.FileField, self).deconstruct()
        if kwargs.get("max_length") == 255:
            del kwargs["max_length"]
        return name, path, args, kwargs

    def generate_filename(self, instance, filename):
        return instance.generate_filename(filename)



import os

from django.core.files import File
from django.core.files.uploadedfile import UploadedFile
from django.utils.translation import gettext_lazy as _

from .utils import cached_method
from .variations import PaperVariation


class TemporaryUploadedFile(UploadedFile):
    """
    Обертка над файлом, удаляющая его при закрытии.

    В отличие от django.core.files.uploadedfile.TemporaryUploadedFile, не создает
    новый временный файл, а оборачивает уже существующий. Используется для передачи
    загруженного файла в функцию сохранения.
    """
    def close(self):
        super().close()
        try:
            os.unlink(self.file.name)
        except OSError:
            pass


class VariationFile(File):
    """
    Файл вариации изображения.
    """

    def __init__(self, instance, variation_name):
        self.instance = instance
        self.variation_name = variation_name
        self.storage = instance.get_file().storage
        filename = self.variation.get_output_filename(instance.name)
        super().__init__(None, filename)

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name
        return self.name == other

    def __hash__(self):
        return hash(self.name)

    def _require_file(self):
        if not self:
            raise ValueError(
                _("Variation '%s' has no file associated with it.") % self.variation_name
            )

    def _get_file(self) -> File:
        self._require_file()
        if getattr(self, "_file", None) is None:
            self._file = self.storage.open(self.name, "rb")
        return self._file

    def _set_file(self, file: File):
        self._file = file

    def _del_file(self):
        del self._file

    file = property(_get_file, _set_file, _del_file)

    @property
    def variation(self) -> PaperVariation:
        variations = self.instance.get_variations()
        return variations[self.variation_name]

    @property
    def path(self) -> str:
        self._require_file()
        return self.storage.path(self.name)

    @property
    def url(self) -> str:
        self._require_file()
        return self.storage.url(self.name)

    @property
    def size(self) -> int:
        self._require_file()
        return self.storage.size(self.name)

    def exists(self) -> bool:
        if not self:
            return False
        return self.storage.exists(self.name)

    def open(self, mode: str = "rb"):
        self._require_file()
        if getattr(self, "_file", None) is None:
            self.file = self.storage.open(self.name, mode)
        else:
            self.file.open(mode)
        return self

    # open() doesn't alter the file's contents, but it does reset the pointer
    open.alters_data = True  # noqa

    def delete(self):
        if not self:
            return

        if hasattr(self, "_file"):
            self.close()
            del self.file

        self.storage.delete(self.name)
        self.name = None

    delete.alters_data = True

    @property
    def closed(self) -> bool:
        file = getattr(self, "_file", None)
        return file is None or file.closed

    def close(self):
        file = getattr(self, "_file", None)
        if file is not None:
            file.close()

    @property
    def width(self) -> int:
        return self._get_image_dimensions()[0]

    @property
    def height(self) -> int:
        return self._get_image_dimensions()[1]

    @cached_method("_dimensions_cache")
    def _get_image_dimensions(self):
        return self.variation.get_output_size(
            (self.instance.width, self.instance.height)
        )
